/*********************************************************************
 *
 *  FM synth to SPI to  MCP4822 dual channel 12-bit DAC
 *
 *********************************************************************
 * Bruce Land Cornell University
 * October 2015
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
/* ====== MCP4822 control word ==============
bit 15 A/B: DACA or DACB Selection bit
1 = Write to DACB
0 = Write to DACA
bit 14 ? Don?t Care
bit 13 GA: Output Gain Selection bit
1 = 1x (VOUT = VREF * D/4096)
0 = 2x (VOUT = 2 * VREF * D/4096), where internal VREF = 2.048V.
bit 12 SHDN: Output Shutdown Control bit
1 = Active mode operation. VOUT is available. ?
0 = Shutdown the selected DAC channel. Analog output is not available at the channel that was shut down.
VOUT pin is connected to 500 k???typical)?
bit 11-0 D11:D0: DAC Input Data bits. Bit x is ignored.
*/
// A-channel, 1x, active
#define DAC_config_chan_A 0b0011000000000000
#define DAC_config_chan_B 0b1011000000000000
#define Fs 20000.0
#define two32 4294967296.0 // 2^32 

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config.h"
// threading library
#include "pt_cornell_1_2.h"
// for sine
#include <math.h>

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_cmd, pt_tick;
// uart control threads
static struct pt pt_input, pt_output, pt_DMA_output ;

// system 1 second interval tick
int sys_time_seconds ;

// === 16:16 fixed point macros ==========================================
typedef signed int fix16 ;
#define multfix16(a,b) ((fix16)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16:16
#define float2fix16(a) ((fix16)((a)*65536.0)) // 2^16
#define fix2float16(a) ((float)(a)/65536.0)
#define fix2int16(a)    ((int)((a)>>16))
#define int2fix16(a)    ((fix16)((a)<<16))
#define divfix16(a,b) ((fix16)((((signed long long)(a)<<16)/(b)))) 
#define sqrtfix16(a) (float2fix16(sqrt(fix2float16(a)))) 
#define absfix16(a) abs(a)
#define onefix16 0x00010000 // int2fix16(1)

//== Timer 2 interrupt handler ===========================================
// actual scaled DAC 
volatile unsigned int DAC_data;
// the DDS units: 1=FM, 2=main frequency
volatile unsigned int phase_accum_fm, phase_incr_fm=1.8*261.0*two32/Fs ;// 
volatile unsigned int phase_accum_main, phase_incr_main=261.0*two32/Fs ;//
volatile SpiChannel spiChn = SPI_CHANNEL2 ;	// the SPI channel to use
volatile int spiClkDiv = 2 ; // 20 MHz max speed for this DAC
// DDS sine table
#define sine_table_size 256
volatile fix16 sine_table[sine_table_size];
// envelope: 1=FM, 2=main frequency
// decay rate, attack rate
volatile fix16 env_fm, wave_fm, dk_fm=float2fix16(0.98), attack_fm=float2fix16(0.0), dk_state_fm, attack_state_fm;
volatile fix16 env_main, wave_main, dk_main=float2fix16(0.995), attack_main=float2fix16(0.0), dk_state_main, attack_state_main;
// fraction of FM sine added as increment to #2 phase accumulator
volatile fix16 fm_depth=float2fix16(1.75);
// time scaling for decay calculation
volatile int dk_interval; // wait some samples between decay calcs
// play flag
volatile int play;
// profiling of ISR
volatile int isr_time;
//=============================
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
{
    mT2ClearIntFlag();
    
    // FM phase
    phase_accum_fm += phase_incr_fm ; 
    // main phase
    phase_accum_main += phase_incr_main + multfix16(sine_table[phase_accum_fm>>24], env_fm) ;
;
     
     // init the exponential decays
    if (play) {
        dk_state_fm = fm_depth; 
        dk_state_main = onefix16; 
        attack_state_fm = fm_depth; 
        attack_state_main = onefix16; 
        play = 0;
        phase_accum_fm = 0;
        phase_accum_main = 0;
        dk_interval = 0;
    }
    
    // envelope calculations are 128 times slower than sample rate
    // computes 4 exponential decays and builds the product envelopes
    if ((dk_interval++ & 0x7f) == 0){
        dk_state_fm = multfix16(dk_state_fm, dk_fm) ;
        dk_state_main = multfix16(dk_state_main, dk_main) ;
        attack_state_fm = multfix16(attack_state_fm, attack_fm);
        attack_state_main = multfix16(attack_state_main, attack_main);
        env_fm = multfix16(fm_depth-attack_state_fm, dk_state_fm) ;
        env_main = multfix16(onefix16-attack_state_main, dk_state_main) ;
    }

    // === Channel A =============
    // CS low to start transaction
     mPORTBClearBits(BIT_0); // start transaction
    // test for ready
     //while (TxBufFullSPI2());
     // write to spi2
     WriteSPI2(DAC_data); 
    
    wave_main = multfix16(sine_table[phase_accum_main>>24], env_main);
    // truncate to 12 bits, read table, convert to int and add offset
    DAC_data = DAC_config_chan_A | (fix2int16(wave_main) + 2048) ; 
    // test for done
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
     // CS high
     mPORTBSetBits(BIT_0); // end transaction
     
     // === Channel B =============
     /*
     // CS low to start transaction
     mPORTBClearBits(BIT_0); // start transaction
    // test for ready
     //while (TxBufFullSPI2());
     // write to spi2
     WriteSPI2(DAC_data);
     
     wave_2 = multfix16(sine_table[phase_accum_2>>24], env_2);
    // truncate to 12 bits, phase-shift 90 degrees, read table, convert to int and add offset
    DAC_data = DAC_config_chan_B | (fix2int16(wave_2) + 2048) ; 
    // test for done
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
     // CS high
     mPORTBSetBits(BIT_0); // end transaction
     */
     isr_time = max(ReadTimer2(), isr_time) ; // - isr_time;
} // end ISR TIMER2

// === Serial Thread ======================================================
static PT_THREAD (protothread_cmd(struct pt *pt))
{
    PT_BEGIN(pt);
    // The serial interface
    static char cmd[16]; 
    static float value;
    // clear port used by thread 4
    mPORTBClearBits(BIT_0);
    
      while(1) {
          
            // send the prompt via DMA to serial
            sprintf(PT_send_buffer,"cmd>");
            // by spawning a print thread
            PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
 
          //spawn a thread to handle terminal input
            // the input thread waits for input
            // -- BUT does NOT block other threads
            // string is returned in "PT_term_buffer"
            PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer(&pt_input) );
            // returns when the thead dies
            // in this case, when <enter> is pushed
            // now parse the string
             sscanf(PT_term_buffer, "%s %f", cmd, &value);

             switch(cmd[0]){
                 case 'f': // frequency 1 is FM; freq 2 is fundamental
                     // freq*2^32/Fs
                     if(cmd[1]=='f') phase_incr_fm = value*(float)two32/Fs;
                     if(cmd[1]=='m') phase_incr_main = value*(float)two32/Fs;
                     break;
                 case 'p': // play
                     play = 1;
                     break;
                 case 'd': // fm depth
                     fm_depth = float2fix16(value);
                     break;
                 case 'k': // decay rate constant 1=FM 2=main
                     if(cmd[1]=='f') dk_fm=float2fix16(value);
                     if(cmd[1]=='m') dk_main=float2fix16(value);
                     break;
                case 'a': // attack rate constant  1=FM 2=main
                     if(cmd[1]=='f') attack_fm=float2fix16(value);
                     if(cmd[1]=='m') attack_main=float2fix16(value);
                     break;
                 case 't':
                     sprintf(PT_send_buffer, "%d\t%d\n\r", isr_time, sys_time_seconds);
                     // by spawning a print thread
                     PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                     break;
             }
             
            // never exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread 3

// === Thread 5 ======================================================
// update a 1 second tick counter
static PT_THREAD (protothread_tick(struct pt *pt))
{
    PT_BEGIN(pt);

      while(1) {
            // yield time 1 second
            PT_YIELD_TIME_msec(1000) ;
            sys_time_seconds++ ;
            // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread 4

// === Main  ======================================================
// set up UART, timer2, threads
// then schedule them as fast as possible

int main(void)
{
  // === config the uart, DMA, vref, timer5 ISR =============
  PT_setup();

   // === setup system wide interrupts  ====================
  INTEnableSystemMultiVectoredInt();
    
  /// timer interrupt //////////////////////////
        // Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
        // at 30 MHz PB clock 60 counts is two microsec
        // 400 is 100 ksamples/sec
        // 2000 is 20 ksamp/sec
        // 1000 is 40 ksample/sec
        // 2000 is 20 ks/sec
        OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 2000);
        // set up the timer interrupt with a priority of 2
        ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
        mT2ClearIntFlag(); // and clear the interrupt flag

        // SCK2 is pin 26 
        // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
        PPSOutput(2, RPB5, SDO2);

        // control CS for DAC
        mPORTBSetPinsDigitalOut(BIT_0);
        mPORTBSetBits(BIT_0);
        
        // divide Fpb by 2, configure the I/O ports. Not using SS in this example
        // 16 bit transfer CKP=1 CKE=1
        // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
        // For any given peripherial, you will need to match these
        SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV , spiClkDiv);
  // === now the threads ====================

  // init the threads
  PT_INIT(&pt_cmd);
  PT_INIT(&pt_tick);

  // build the sine lookup table
   // scaled to produce values between 0 and 4096
   int i;
   for (i = 0; i < sine_table_size; i++){
         sine_table[i] =  float2fix16(2047*sin((float)i*6.283/(float)sine_table_size));
    }
        
  // schedule the threads
  while(1) {
    // round robin
    PT_SCHEDULE(protothread_cmd(&pt_cmd));
    PT_SCHEDULE(protothread_tick(&pt_tick));
  }
} // main
